package algs31;

import stdlib.StdOut;
import stdlib.StdRandom;
import stdlib.Stopwatch;

/**
 * Performance test demo for different implementations of symbol tables.
 * The demo uses Integers for keys and Strings (the English spelling of
 * an integer value) for values.
 *
 * Run the test separately for the different implementations to see the behavior.
 *
 * @author Eric Fredericks
 */
public class XSTTest {
    enum Implementation {
        UNORDERED_ARRAY, ORDERED_ARRAY, UNORDERED_LINKED_LIST, BINARY_SEARCH_TREE
    }

    public static void main(String[] args) {
        int MIN = 256;
        int MAX = 100_000_000;

        // TODO: Change this to see the performance of different implementations of symbol tables.
        Implementation implementation = Implementation.BINARY_SEARCH_TREE;

        StdOut.format("\nTotal time for N symbol table operations with implementation: %s\n\n", implementation.toString());
        SymbolTable<Integer, String> st = newSymbolTable(implementation, MIN);
        double prevPutTime = timeTrialPut(st, MIN);
        double prevGetTime = timeTrialGet(st, MIN);
        double prevContainsTime = timeTrialContains(st, MIN);
        double prevDeleteTime = timeTrialDelete(st, MIN);
        for (int N = MIN*2; N<=MAX; N += N) {
            st = newSymbolTable(implementation, N);
            StdOut.format("Elapsed count N=%,13d: | ", N);

            double putTime = timeTrialPut(st, N);
            StdOut.format("Put[%10.3f : %10.3f] | ", putTime, putTime/prevPutTime);

            double getTime = timeTrialGet(st, N);
            StdOut.format("Get[%10.3f : %10.3f] | ", getTime, getTime/prevGetTime);

            double containsTime = timeTrialContains(st, N);
            StdOut.format("Contains[%10.3f : %10.3f] | ", containsTime, containsTime/prevContainsTime);

            double deleteTime = timeTrialDelete(st, N);
            StdOut.format("Delete[%10.3f : %10.3f]\n", deleteTime, deleteTime/prevDeleteTime);

            prevGetTime = getTime;
            prevPutTime = putTime;
            prevContainsTime = containsTime;
            prevDeleteTime = deleteTime;
        }
    }

    private static SymbolTable<Integer, String> newSymbolTable(Implementation implementation, int capacity) {
        return switch (implementation) {
            case UNORDERED_ARRAY -> new ArrayST<>(capacity); // eliminate array resizing
            case ORDERED_ARRAY -> new BinarySearchST<>(capacity); // eliminate array resizing
            case UNORDERED_LINKED_LIST -> new SequentialSearchST<>();
            case BINARY_SEARCH_TREE -> new BinarySearchTreeST<>();
        };
    }

    private static double timeTrialPut(SymbolTable<Integer, String> st, int N) {
        if(!st.isEmpty()) throw new IllegalStateException("Symbol table for put must be empty");
        int[] a = new int[N];
        for (int j=0; j<N; j+=1) {
            a[j] = j;
        }

        // VERY IMPORTANT: shuffle the inputs!
        StdRandom.shuffle(a);

        Stopwatch s = new Stopwatch();
        for (int j=0; j<N; j+=1) {
            st.put(a[j], numberToWords(a[j]));
        }
        return s.elapsedTime();
    }

    private static double timeTrialGet(SymbolTable<Integer, String> st, int N) {
        if(st.size() != N) throw new IllegalStateException("Symbol table for get must have N elements");
        Stopwatch s = new Stopwatch();
        for (int j=0; j<N; j+=1) {
            st.get(j);
        }
        return s.elapsedTime();
    }

    private static double timeTrialContains(SymbolTable<Integer, String> st,int N) {
        if(st.size() != N) throw new IllegalStateException("Symbol table for contains must have N elements");
        Stopwatch s = new Stopwatch();
        for (int j=0; j<N; j+=1) {
            st.contains(j);
        }
        return s.elapsedTime();
    }

    private static double timeTrialDelete(SymbolTable<Integer, String> st,int N) {
        if(st.size() != N) throw new IllegalStateException("Symbol table for delete must have N elements");
        Stopwatch s = new Stopwatch();
        for (int j=0; j<N; j+=1) {
            st.delete(j);
        }
        return s.elapsedTime();
    }

    // Adapted from https://walkccc.me/LeetCode/problems/0273/#__tabbed_1_2
    public static String numberToWords(int num) {
        return num == 0 ? "Zero" : helper(num);
    }

    private static final String[] belowTwenty = {"", "One", "Two", "Three", "Four",
            "Five", "Six", "Seven", "Eight", "Nine",
            "Ten", "Eleven", "Twelve", "Thirteen", "Fourteen",
            "Fifteen", "Sixteen", "Seventeen", "Eighteen", "Nineteen"};
    private static final String[] tens = {"", "", "Twenty", "Thirty", "Forty",
            "Fifty", "Sixty", "Seventy", "Eighty", "Ninety"};

    private static String helper(int num) {
        StringBuilder s = new StringBuilder();

        if (num < 20)
            s.append(belowTwenty[num]);
        else if (num < 100)
            s.append(tens[num / 10]).append(" ").append(belowTwenty[num % 10]);
        else if (num < 1000)
            s.append(helper(num / 100)).append(" Hundred ").append(helper(num % 100));
        else if (num < 1000000)
            s.append(helper(num / 1000)).append(" Thousand ").append(helper(num % 1000));
        else if (num < 1000000000)
            s.append(helper(num / 1_000_000)).append(" Million ").append(helper(num % 1_000_000));
        else
            s.append(helper(num / 1_000_000_000)).append(" Billion ").append(helper(num % 1_000_000_000));

        return s.toString().trim();
    }
}
